def breadth_first_search(graph, start_node):
    """

    Example:
    >>> graph = {
    >>>     'A': ['B', 'C'],
    >>>     'B': ['D', 'E'],
    >>>     'C': ['F', 'G'],
    >>>     'D': [],
    >>>     'E': [],
    >>>     'F': [],
    >>>     'G': []
    >>> }
    >>> breadth_first_search(graph, 'A')
    A B C D E F G

    Code:
    -----
    def breadth_first_search(graph, start_node):
        visited = []  
        queue = []    

        visited.append(start_node)
        queue.append(start_node)

        while queue:
            current_node = queue.pop(0)
            print(current_node, end=" ")
            for neighbor in graph[current_node]:
                if neighbor not in visited:
                    visited.append(neighbor)
                    queue.append(neighbor)
    """
    
    visited = []  
    queue = []    

    visited.append(start_node)
    queue.append(start_node)

    while queue:
        current_node = queue.pop(0)
        print(current_node, end=" ")
        for neighbor in graph[current_node]:
            if neighbor not in visited:
                visited.append(neighbor)
                queue.append(neighbor)




def depth_first_search(graph, start_node):
    """
    Example:
    >>> graph = {
    >>>     'A': ['B', 'C'],
    >>>     'B': ['D', 'E'],
    >>>     'C': ['F', 'G'],
    >>>     'D': [],
    >>>     'E': [],
    >>>     'F': [],
    >>>     'G': []
    >>> }
    >>> depth_first_search(graph, 'A')
    A B D E C F G

    Code:
    -----
    def depth_first_search(graph, start_node):
        visited = set() 

        def dfs(node):
            if node not in visited:
                print(node, end=" ")
                visited.add(node)
                for neighbor in graph[node]:
                    dfs(neighbor)

        dfs(start_node)
    """
    
    visited = set() 

    def dfs(node):
        if node not in visited:
            print(node, end=" ")
            visited.add(node)
            for neighbor in graph[node]:
                dfs(neighbor)

    dfs(start_node)




def map_coloring_csp():
    """

    Example:
    >>> result = map_coloring_csp()
    >>> print(result)

    Code:
    -----
    def map_coloring_csp():
        variables = ['WA', 'NT', 'SA', 'Q', 'NSW', 'V', 'T']


        domains = {}
        for var in variables:
            domains[var] = ['red', 'green', 'blue']


        neighbors = {
            'WA': ['NT', 'SA'],
            'NT': ['WA', 'Q', 'SA'],
            'SA': ['WA', 'NT', 'Q', 'NSW', 'V'],
            'Q': ['NT', 'SA', 'NSW'],
            'NSW': ['Q', 'SA', 'V'],
            'V': ['SA', 'NSW'],
            'T': []
        }


        def backtracking_search(assignment):
            if len(assignment) == len(variables):
                return assignment

            var = select_unassigned_variable(assignment)
            for value in order_domain_values(var, assignment):
                if is_consistent(var, value, assignment):
                    assignment[var] = value
                    result = backtracking_search(assignment)
                    if result is not None:
                        return result
                    assignment.pop(var)

            return None


        def select_unassigned_variable(assignment):
            for var in variables:
                if var not in assignment:
                    return var


        def order_domain_values(var, assignment):
            return domains[var]

        def is_consistent(var, value, assignment):
            for neighbor in neighbors[var]:
                if neighbor in assignment and assignment[neighbor] == value:
                    return False
            return True

        assignment = {}
        result = backtracking_search(assignment)

        return result
    """


    variables = ['WA', 'NT', 'SA', 'Q', 'NSW', 'V', 'T']


    domains = {}
    for var in variables:
        domains[var] = ['red', 'green', 'blue']


    neighbors = {
        'WA': ['NT', 'SA'],
        'NT': ['WA', 'Q', 'SA'],
        'SA': ['WA', 'NT', 'Q', 'NSW', 'V'],
        'Q': ['NT', 'SA', 'NSW'],
        'NSW': ['Q', 'SA', 'V'],
        'V': ['SA', 'NSW'],
        'T': []
    }


    def backtracking_search(assignment):
        if len(assignment) == len(variables):
            return assignment

        var = select_unassigned_variable(assignment)
        for value in order_domain_values(var, assignment):
            if is_consistent(var, value, assignment):
                assignment[var] = value
                result = backtracking_search(assignment)
                if result is not None:
                    return result
                assignment.pop(var)

        return None


    def select_unassigned_variable(assignment):
        for var in variables:
            if var not in assignment:
                return var


    def order_domain_values(var, assignment):
        return domains[var]

    def is_consistent(var, value, assignment):
        for neighbor in neighbors[var]:
            if neighbor in assignment and assignment[neighbor] == value:
                return False
        return True

    assignment = {}
    result = backtracking_search(assignment)

    return result




# a_star.py

from queue import PriorityQueue

def a_star_search(graph, start, goal, heuristic):
    """
    Usage:
    >>> graph = {
    >>>     'S': [('A', 1), ('B', 4)],
    >>>     'A': [('C', 2)],
    >>>     'B': [('D', 2)],
    >>>     'C': [('G', 3)],
    >>>     'D': [('G', 1)],
    >>>     'G': []
    >>> }
    >>> heuristic = {
    >>>     'S': 7,
    >>>     'A': 6,
    >>>     'B': 2,
    >>>     'C': 1,
    >>>     'D': 2,
    >>>     'G': 0
    >>> }
    >>> path = a_star_search(graph, 'S', 'G', heuristic)
    >>> print("Path:", " -> ".join(path))

    Code:
    -----
    from queue import PriorityQueue

    def a_star_search(graph, start, goal, heuristic):
        open_list = PriorityQueue()
        open_list.put((0, start))  
        came_from = {}  
        g_score = {start: 0}  
        f_score = {start: heuristic[start]}  
        
        while not open_list.empty():
            current = open_list.get()[1]  
            if current == goal:

                path = []
                while current in came_from:
                    path.append(current)
                    current = came_from[current]
                path.append(start)
                return path[::-1]  
            
            for neighbor, cost in graph[current]:
                temp_g_score = g_score[current] + cost  
                
                if neighbor not in g_score or temp_g_score < g_score[neighbor]:
                    g_score[neighbor] = temp_g_score
                    f_score[neighbor] = g_score[neighbor] + heuristic[neighbor]
                    open_list.put((f_score[neighbor], neighbor))
                    came_from[neighbor] = current  

        return [] 
    """
    open_list = PriorityQueue()
    open_list.put((0, start))  
    came_from = {}  
    g_score = {start: 0}  
    f_score = {start: heuristic[start]}  
    
    while not open_list.empty():
        current = open_list.get()[1]  
        if current == goal:

            path = []
            while current in came_from:
                path.append(current)
                current = came_from[current]
            path.append(start)
            return path[::-1]  
        
        for neighbor, cost in graph[current]:
            temp_g_score = g_score[current] + cost  
            
            if neighbor not in g_score or temp_g_score < g_score[neighbor]:
                g_score[neighbor] = temp_g_score
                f_score[neighbor] = g_score[neighbor] + heuristic[neighbor]
                open_list.put((f_score[neighbor], neighbor))
                came_from[neighbor] = current  

    return []  




def minimax_with_alpha_beta(depth, nodeIndex, maximizingPlayer, values, alpha, beta):
    """
    Example:
    >>> values = [3, 5, 2, 9, 12, 15, 7, 8]
    >>> print("The optimal value is:", minimax_with_alpha_beta(0, 0, True, values, float('-inf'), float('inf')))
    The optimal value is: 12

    Code:
    -----
    def minimax_with_alpha_beta(depth, nodeIndex, maximizingPlayer, values, alpha, beta):
        if depth == 3:
            return values[nodeIndex]
        
        if maximizingPlayer:
            best = float('-inf')
            for i in range(2):
                val = minimax_with_alpha_beta(depth + 1, nodeIndex * 2 + i, False, values, alpha, beta)
                best = max(best, val)
                alpha = max(alpha, best)
                if beta <= alpha:
                    break
            return best
        else:
            best = float('inf')
            for i in range(2):
                val = minimax_with_alpha_beta(depth + 1, nodeIndex * 2 + i, True, values, alpha, beta)
                best = min(best, val)
                beta = min(beta, best)
                if beta <= alpha:
                    break
            return best
    """
    
    if depth == 3:
        return values[nodeIndex]
    
    if maximizingPlayer:
        best = float('-inf')
        for i in range(2):
            val = minimax_with_alpha_beta(depth + 1, nodeIndex * 2 + i, False, values, alpha, beta)
            best = max(best, val)
            alpha = max(alpha, best)
            if beta <= alpha:
                break
        return best
    else:
        best = float('inf')
        for i in range(2):
            val = minimax_with_alpha_beta(depth + 1, nodeIndex * 2 + i, True, values, alpha, beta)
            best = min(best, val)
            beta = min(beta, best)
            if beta <= alpha:
                break
        return best




def alpha_beta_pruning(depth, index, maximizing_player, values, alpha, beta):
    """
   
    Example:
    >>> values = [3, 5, 6, 9, 1, 2, 0, -1]
    >>> print(alpha_beta_pruning(0, 0, True, values, float('-inf'), float('inf')))
    5

    code:
    -----
    def alpha_beta_pruning(depth, index, maximizing_player, values, alpha, beta):
        if depth == 3:
            return values[index]
        
        if maximizing_player:
            best = float('-inf')
            for j in range(2):
                val = alpha_beta_pruning(depth + 1, index * 2 + j, False, values, alpha, beta)
                best = max(best, val)
                alpha = max(alpha, best)
                if beta <= alpha:
                    break
            return best
        else:
            best = float('inf')
            for j in range(2):
                val = alpha_beta_pruning(depth + 1, index * 2 + j, True, values, alpha, beta)
                best = min(best, val)
                beta = min(beta, best)
                if beta <= alpha:
                    break
            return best
    """
    if depth == 3:
        return values[index]
    
    if maximizing_player:
        best = float('-inf')
        for j in range(2):
            val = alpha_beta_pruning(depth + 1, index * 2 + j, False, values, alpha, beta)
            best = max(best, val)
            alpha = max(alpha, best)
            if beta <= alpha:
                break
        return best
    else:
        best = float('inf')
        for j in range(2):
            val = alpha_beta_pruning(depth + 1, index * 2 + j, True, values, alpha, beta)
            best = min(best, val)
            beta = min(beta, best)
            if beta <= alpha:
                break
        return best





# Import search-related algorithms
from .a_star import a_star_search
from .bfs import breadth_first_search
from .dfs import depth_first_search
from .gbfs import greedy_best_first_search
from .uniform_cost_search import uniform_cost_search
from .memory_bounded_heuristic import ida_star
from .iterative_deepening_search import iterative_deepening_search

# Import AI algorithms
from .alpha_beta import alpha_beta_pruning
from .minmax import minimax_with_alpha_beta
from .dpll_algo import dpll_algorithm

# Import knowledge-based reasoning algorithms
from .backward_chaining import BackwardKB 
from .forward_chaining import ForwardKB 

# Import constraint satisfaction and other problem-solving algorithms
from .constraint_satisfaction_problem import map_coloring_csp
from .n_queens import n_queens_solver

# Import other miscellaneous algorithms
from .eight_puzzle import solve_puzzle
from .bayesian_network import bayesian_network
from .cryptarithmetic import cryptarithmetic



def iterative_deepening_search(start_state, goal):
    """
    Example:
    >>>    start_state = "A"  # Starting node
    >>>    goal = "G"         # Goal node
    >>>    path = iterative_deepening_search(start_state, goal)
    >>>    if path:
    >>>        print("Path to goal:", path)
    >>>    else:
    >>>         print("Goal not found")


    code:
    ----
    class Node:
        def __init__(self, state, parent=None):
            self.state = state
            self.parent = parent
    
    def depth_limited_search(start, goal, depth):
        if start.state == goal:
            return start
        elif depth == 0:
            return None
        else:
            for child_state in get_children(start.state):
                child = Node(child_state, start)
                result = depth_limited_search(child, goal, depth - 1)
                if result is not None:
                    return result
        return None
    
    def get_children(state):
        graph = {
            "A": ["B", "C"],
            "B": ["D", "E"],
            "C": ["F", "G"],
            "D": [],
            "E": [],
            "F": [],
            "G": []
        }
        return graph.get(state, [])
    
    # Perform iterative deepening
    depth = 0
    while True:
        result = depth_limited_search(Node(start_state), goal, depth)
        if result is not None:
            # Reconstruct the path
            path = []
            while result is not None:
                path.append(result.state)
                result = result.parent
            path.reverse()
            return path  # Path to the goal
        depth += 1

    """
    
    class Node:
        def __init__(self, state, parent=None):
            self.state = state
            self.parent = parent
    
    def depth_limited_search(start, goal, depth):
        if start.state == goal:
            return start
        elif depth == 0:
            return None
        else:
            for child_state in get_children(start.state):
                child = Node(child_state, start)
                result = depth_limited_search(child, goal, depth - 1)
                if result is not None:
                    return result
        return None
    
    def get_children(state):
        graph = {
            "A": ["B", "C"],
            "B": ["D", "E"],
            "C": ["F", "G"],
            "D": [],
            "E": [],
            "F": [],
            "G": []
        }
        return graph.get(state, [])
    
    # Perform iterative deepening
    depth = 0
    while True:
        result = depth_limited_search(Node(start_state), goal, depth)
        if result is not None:
            # Reconstruct the path
            path = []
            while result is not None:
                path.append(result.state)
                result = result.parent
            path.reverse()
            return path  # Path to the goal
        depth += 1




class BackwardKB:
    """
    Example:
    >>> kb = BackwardKB()
    >>> kb.add_fact("A")
    >>> kb.add_fact("B")
    >>> kb.add_rule({"if": {"A", "B"}, "then": "C"})
    >>> kb.add_rule({"if": {"C"}, "then": "D"})
    >>> goal = "D"
    >>> print(f"Goal {goal} is {'achieved' if kb.backward_chaining(goal) else 'not achievable'}.")
    Goal D is achieved.

    Code:
    class BackwardKB:
        def __init__(self):
            self.rules = []
            self.facts = set()

        def add_rule(self, rule):
        self.rules.append(rule)

        def add_fact(self, fact):
            self.facts.add(fact)

        def backward_chaining(self, goal):
            if goal in self.facts:
                return True
            for rule in self.rules:
                if rule["then"] == goal and all(self.backward_chaining(g) for g in rule["if"]):
                    self.facts.add(goal)
                    return True
            return False
        
    """

    def __init__(self):
        self.rules = []
        self.facts = set()

    def add_rule(self, rule):
       self.rules.append(rule)

    def add_fact(self, fact):
        self.facts.add(fact)

    def backward_chaining(self, goal):
        if goal in self.facts:
            return True
        for rule in self.rules:
            if rule["then"] == goal and all(self.backward_chaining(g) for g in rule["if"]):
                self.facts.add(goal)
                return True
        return False





def ida_star(start_state, goal_state, heuristic_func):
    """
    
    Example:
    >>>    def heuristic_func(state):
    >>>        heuristic_values = {
    >>>            'A': 7, 'B': 6, 'C': 2, 'D': 1, 'E': 3, 'F': 4, 'G': 0, 
    >>>            'H': 5, 'I': 8, 'J': 9
    >>>        }
    >>>        return heuristic_values.get(state, 0)

    >>>    start_state = 'A'
    >>>    goal_state = 'G'
    >>>    result = ida_star(start_state, goal_state, heuristic_func)

    >>>    if result:
    >>>        path = []
    >>>        current = result
    >>>        while current:
    >>>            path.append(current.state)
    >>>            current = current.parent
    >>>        print("Path to goal:", " -> ".join(path[::-1]))
    >>>        print("Total cost:", result.cost)
    >>>    else:
    >>>        print("Goal not reachable.")


    Code:
    def ida_star(start_state, goal_state, heuristic_func):
        class Node:
            def __init__(self, state, parent, cost, heuristic):
                self.state = state
                self.parent = parent
                self.cost = cost
                self.heuristic = heuristic
                self.f = cost + heuristic

        def get_successors(state):
            successors = {
                'A': [('B', 3), ('C', 6)],
                'B': [('D', 2), ('E', 4)],
                'C': [('F', 1), ('G', 5)],
                'D': [('H', 7)],
                'E': [('I', 8)],
                'F': [('J', 9)],
                'G': [],
                'H': [], 'I': [], 'J': []
            }
            return successors.get(state, [])

        def search(path, g, threshold):
            node = path[-1]
            f = g + heuristic_func(node.state)
            if f > threshold:
                return f
            if node.state == goal_state:
                return node
            min_threshold = float('inf')
            for state, cost in get_successors(node.state):
                if state not in [n.state for n in path]:
                    successor = Node(state, node, g + cost, heuristic_func(state))
                    path.append(successor)
                    temp = search(path, g + cost, threshold)
                    if isinstance(temp, Node):
                        return temp
                    min_threshold = min(min_threshold, temp)
                    path.pop()
            return min_threshold

        threshold = heuristic_func(start_state)
        start_node = Node(start_state, None, 0, heuristic_func(start_state))
        path = [start_node]
        while True:
            temp = search(path, 0, threshold)
            if isinstance(temp, Node):
                return temp
            if temp == float('inf'):
                return None
            threshold = temp

    """
    
    class Node:
        def __init__(self, state, parent, cost, heuristic):
            self.state = state
            self.parent = parent
            self.cost = cost
            self.heuristic = heuristic
            self.f = cost + heuristic

    def get_successors(state):
        successors = {
            'A': [('B', 3), ('C', 6)],
            'B': [('D', 2), ('E', 4)],
            'C': [('F', 1), ('G', 5)],
            'D': [('H', 7)],
            'E': [('I', 8)],
            'F': [('J', 9)],
            'G': [],
            'H': [], 'I': [], 'J': []
        }
        return successors.get(state, [])

    def search(path, g, threshold):
        node = path[-1]
        f = g + heuristic_func(node.state)
        if f > threshold:
            return f
        if node.state == goal_state:
            return node
        min_threshold = float('inf')
        for state, cost in get_successors(node.state):
            if state not in [n.state for n in path]:
                successor = Node(state, node, g + cost, heuristic_func(state))
                path.append(successor)
                temp = search(path, g + cost, threshold)
                if isinstance(temp, Node):
                    return temp
                min_threshold = min(min_threshold, temp)
                path.pop()
        return min_threshold

    threshold = heuristic_func(start_state)
    start_node = Node(start_state, None, 0, heuristic_func(start_state))
    path = [start_node]
    while True:
        temp = search(path, 0, threshold)
        if isinstance(temp, Node):
            return temp
        if temp == float('inf'):
            return None
        threshold = temp





import heapq

def uniform_cost_search(graph, start, goal):
    """

    Example:
    >>> graph = {'S': [('A', 1), ('B', 4)], 'A': [('C', 2), ('D', 5)], 'B': [('D', 2)], 'C': [('G', 3)], 'D': [('G', 2)], 'G': []}
    >>> uniform_cost_search(graph, 'S', 'G')
    (5, ['S', 'A', 'C', 'G'])


    Code:
    -----
    import heapq

    def uniform_cost_search(graph, start, goal):
        pq, visited = [(0, start, [])], set()

        while pq:
            cost, node, path = heapq.heappop(pq)

            if node in visited:
                continue
            
            visited.add(node)
            path = path + [node]

            if node == goal:
                return cost, path
            
            for neighbor, edge_cost in graph.get(node, []):
                if neighbor not in visited:
                    heapq.heappush(pq, (cost + edge_cost, neighbor, path))

        return float("inf"), []

    """
    pq, visited = [(0, start, [])], set()

    while pq:
        cost, node, path = heapq.heappop(pq)

        if node in visited:
            continue
        
        visited.add(node)
        path = path + [node]

        if node == goal:
            return cost, path
        
        for neighbor, edge_cost in graph.get(node, []):
            if neighbor not in visited:
                heapq.heappush(pq, (cost + edge_cost, neighbor, path))

    return float("inf"), []



from pgmpy.models import BayesianNetwork
from pgmpy.factors.discrete import TabularCPD
from pgmpy.inference import BeliefPropagation

def bayesian_network():
    """
    Example:
    >>> bayesian_network()
    Model is valid: True
    +------+----------+
    | C    |   phi(C) |
    +------+----------+
    | C(0) |   0.6000 |
    | C(1) |   0.4000 |
    +------+----------+

    Code:
    ----
    from pgmpy.models import BayesianNetwork
    from pgmpy.factors.discrete import TabularCPD
    from pgmpy.inference import BeliefPropagation

    def bayesian_network():
        model = BayesianNetwork([('A', 'C'), ('B', 'C')])

        model.add_cpds(
            TabularCPD('A', 2, [[0.8], [0.2]]),
            TabularCPD('B', 2, [[0.7], [0.3]]),
            TabularCPD('C', 2, [[0.9, 0.6, 0.7, 0.1], [0.1, 0.4, 0.3, 0.9]], evidence=['A', 'B'], evidence_card=[2, 2])
        )

        print("Model is valid:", model.check_model())

        infer = BeliefPropagation(model)
        result = infer.query(variables=['C'], evidence={'A': 1, 'B': 0})

        print(result)
    """
    
    
    model = BayesianNetwork([('A', 'C'), ('B', 'C')])

    model.add_cpds(
        TabularCPD('A', 2, [[0.8], [0.2]]),
        TabularCPD('B', 2, [[0.7], [0.3]]),
        TabularCPD('C', 2, [[0.9, 0.6, 0.7, 0.1], [0.1, 0.4, 0.3, 0.9]], evidence=['A', 'B'], evidence_card=[2, 2])
    )

    print("Model is valid:", model.check_model())

    infer = BeliefPropagation(model)
    result = infer.query(variables=['C'], evidence={'A': 1, 'B': 0})

    print(result)





def dpll_algorithm(clauses, assignment):
    """
    
    Example:
    >>> clauses = [['A', '¬B'], ['¬A', 'B'], ['¬A']]
    >>> assignment = {}
    >>> print(dpll_algorithm(clauses, assignment))
    {'A': True, '¬B': True, '¬A': True}


    Code:
    -----
    def dpll_algorithm(clauses, assignment):
        def is_satisfied(clause, assignment):

            for literal in clause:
                if literal in assignment and assignment[literal]:
                    return True
            return False

        def is_unsatisfied(clause, assignment):
            for literal in clause:
                if literal not in assignment:
                    return False
            return all(not assignment[literal] for literal in clause)

        def choose_unassigned_variable(clauses, assignment):
            for clause in clauses:
                for literal in clause:
                    if literal not in assignment:
                        return literal

        if all(is_satisfied(clause, assignment) for clause in clauses):
            return assignment

        if any(is_unsatisfied(clause, assignment) for clause in clauses):
            return False

        var = choose_unassigned_variable(clauses, assignment)

        assignment[var] = True
        result = dpll_algorithm(clauses, assignment)
        if result:
            return result

        assignment[var] = False
        return dpll_algorithm(clauses, assignment)

    """

    def is_satisfied(clause, assignment):

        for literal in clause:
            if literal in assignment and assignment[literal]:
                return True
        return False

    def is_unsatisfied(clause, assignment):
        for literal in clause:
            if literal not in assignment:
                return False
        return all(not assignment[literal] for literal in clause)

    def choose_unassigned_variable(clauses, assignment):
        for clause in clauses:
            for literal in clause:
                if literal not in assignment:
                    return literal

    if all(is_satisfied(clause, assignment) for clause in clauses):
        return assignment

    if any(is_unsatisfied(clause, assignment) for clause in clauses):
        return False

    var = choose_unassigned_variable(clauses, assignment)

    assignment[var] = True
    result = dpll_algorithm(clauses, assignment)
    if result:
        return result

    assignment[var] = False
    return dpll_algorithm(clauses, assignment)




from itertools import permutations

def cryptarithmetic():
    """

    
    Example:
    >>> cryptarithmetic()
    SEND = 9567
    MORE = 1085
    MONEY = 10652

    Code:
    -----
    def cryptarithmetic():
        letters = "SENDMORY"
        digits = range(10)

        for perm in permutations(digits, len(letters)):
            s, e, n, d, m, o, r, y = perm

            if s == 0 or m == 0:
                continue

            send = s * 1000 + e * 100 + n * 10 + d
            more = m * 1000 + o * 100 + r * 10 + e
            money = m * 10000 + o * 1000 + n * 100 + e * 10 + y

            if send + more == money:
                print("SEND =", send)
                print("MORE =", more)
                print("MONEY =", money)
                break
    """
    
    letters = "SENDMORY"
    digits = range(10)

    for perm in permutations(digits, len(letters)):
        s, e, n, d, m, o, r, y = perm

        if s == 0 or m == 0:
            continue

        send = s * 1000 + e * 100 + n * 10 + d
        more = m * 1000 + o * 100 + r * 10 + e
        money = m * 10000 + o * 1000 + n * 100 + e * 10 + y

        if send + more == money:
            print("SEND =", send)
            print("MORE =", more)
            print("MONEY =", money)
            break





def n_queens_solver(n):
    """

    
    Example:
    >>> n_queens_solver(4)
    Enter the number of queens: 4
    Number of solutions: 2
    Solution 1:
    . Q . .
    . . . Q
    Q . . .
    . . Q .

    Solution 2:
    . . Q .
    Q . . .
    . . . Q
    . Q . .


    Code:
    -----
    def n_queens_solver(n):
        def print_solution(board):
            for row in board:
                print(" ".join("Q" if cell else "." for cell in row))
            print()

        def is_safe(board, row, col, n):
            return not any(
                board[i][col] or 
                (col - (row - i) >= 0 and board[i][col - (row - i)]) or
                (col + (row - i) < n and board[i][col + (row - i)])
                for i in range(row)
            )

        def solve_n_queens(board, row, n, solutions):
            if row == n:
                solutions.append([row[:] for row in board])  # Found a solution
            else:
                for col in range(n):
                    if is_safe(board, row, col, n):
                        board[row][col] = 1  # Place queen
                        solve_n_queens(board, row + 1, n, solutions)  # Recurse to place queens in next row
                        board[row][col] = 0  # Backtrack

        solutions = []
        solve_n_queens([[0] * n for _ in range(n)], 0, n, solutions)

        print(f"Number of solutions: {len(solutions)}")
        for index, solution in enumerate(solutions, start=1):
            print(f"Solution {index}:")
            print_solution(solution)

        return solutions
    """
    
    def print_solution(board):
        for row in board:
            print(" ".join("Q" if cell else "." for cell in row))
        print()

    def is_safe(board, row, col, n):
        return not any(
            board[i][col] or 
            (col - (row - i) >= 0 and board[i][col - (row - i)]) or
            (col + (row - i) < n and board[i][col + (row - i)])
            for i in range(row)
        )

    def solve_n_queens(board, row, n, solutions):
        if row == n:
            solutions.append([row[:] for row in board])  # Found a solution
        else:
            for col in range(n):
                if is_safe(board, row, col, n):
                    board[row][col] = 1  # Place queen
                    solve_n_queens(board, row + 1, n, solutions)  # Recurse to place queens in next row
                    board[row][col] = 0  # Backtrack

    solutions = []
    solve_n_queens([[0] * n for _ in range(n)], 0, n, solutions)

    print(f"Number of solutions: {len(solutions)}")
    for index, solution in enumerate(solutions, start=1):
        print(f"Solution {index}:")
        print_solution(solution)

    return solutions



# greedy_best_first_search.py

import networkx as nx

def greedy_best_first_search(graph, start, goal, heuristic):
    """


    Usage:
    >>> graph = nx.Graph()
    >>> graph.add_weighted_edges_from([
    >>>     ('S', 'A', 1), ('S', 'B', 4),
    >>>     ('A', 'C', 2), ('B', 'D', 2),
    >>>     ('C', 'G', 2), ('D', 'G', 3)
    >>> ])
    >>> heuristic = {'S': 7, 'A': 6, 'B': 2, 'C': 1, 'D': 2, 'G': 0}
    >>> path = greedy_best_first_search(graph, 'S', 'G', heuristic)
    >>> print("Path:", " -> ".join(path))
    Path: S -> B -> D -> G

    Code:
    -----
  
  import networkx as nx

  def greedy_best_first_search(graph, start, goal, heuristic):
    path = nx.astar_path(graph, start, goal, heuristic=lambda n, _: heuristic[n])
    return path
    """
    
    path = nx.astar_path(graph, start, goal, heuristic=lambda n, _: heuristic[n])
    return path



class ForwardKB:
    """
    Example:
   >>> kb = ForwardKB()
   >>> kb.add_fact("A")
   >>> kb.add_fact("B")
   >>> kb.add_rule({"if": {"A", "B"}, "then": "C"})
   >>> kb.add_rule({"if": {"C"}, "then": "D"})
   >>> kb.forward_chaining()
   >>> print(f"Final facts: {kb.facts}")
    # Output: Inferred: C
    #         Inferred: D
    # Final facts will include: {'A', 'B', 'C', 'D'}
     |
Code:
----
     |
class ForwardKB:
    def __init__(self):
        self.rules = []
        self.facts = set()
     |
    def add_rule(self, rule):
        self.rules.append(rule)
     |
    def add_fact(self, fact):
        self.facts.add(fact)
     |
    def forward_chaining(self):
        while True:
            new_inference = False
            for rule in self.rules:
                if rule["if"].issubset(self.facts) and rule["then"] not in self.facts:
                    self.facts.add(rule["then"])
                    print(f"Inferred: {rule['then']}")
                    new_inference = True
            if not new_inference:
  
    """

    def __init__(self):
        self.rules = []
        self.facts = set()

    def add_rule(self, rule):
        self.rules.append(rule)

    def add_fact(self, fact):
        self.facts.add(fact)

    def forward_chaining(self):
        while True:
            new_inference = False
            for rule in self.rules:
                if set(rule["if"]).issubset(self.facts) and rule["then"] not in self.facts:
                    self.facts.add(rule["then"])
                    print(f"Inferred: {rule['then']}")
                    new_inference = True
            if not new_inference:
                break  # End the loop when no new facts are inferred



from setuptools import setup, find_packages

setup(
    name='algoai',  
    version='0.2.3',
    packages=find_packages(),
    install_requires=[  
        'numpy',   
        'networkx', 
        'pgmpy',
        
    ],
    author='Leander',
    author_email='leander.antony2023@gmail.com',
    description='A collection of AI algorithms',
    long_description=open('README.md').read(), 
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6', 
)





from heapq import heappop, heappush

def solve_puzzle(start, goal):
    """


    Usage:
    >>> start = (1, 2, 3, 4, 0, 5, 6, 7, 8)
    >>> goal = (1, 2, 3, 4, 5, 6, 7, 8, 0)
    >>> steps, path = solve_puzzle(start, goal)
    >>> print(f"Steps: {steps}\nPath: {path}")
    Steps: 5
    Path: ((1, 2, 3, 4, 0, 5, 6, 7, 8), (1, 2, 3, 4, 5, 0, 6, 7, 8), ...)

    Code:
    -----
    from heapq import heappop, heappush

    def solve_puzzle(start, goal):
        def h(state):
            return sum(abs((v-1) % 3 - i % 3) + abs((v-1) // 3 - i // 3) for i, v in enumerate(state) if v != 0)

        def n(state):
            idx = state.index(0)  
            neighbors = []
            for d in (-1, 1, -3, 3):  
                new_idx = idx + d
                if 0 <= new_idx < 9:

                    if (d == -1 and idx % 3 == 0) or (d == 1 and idx % 3 == 2):
                        continue

                    new_state = list(state)
                    new_state[idx], new_state[new_idx] = new_state[new_idx], new_state[idx]
                    neighbors.append(tuple(new_state))
            return neighbors


        q, visited = [(h(start), 0, start, ())], set()

        while q:
            _, cost, current_state, path = heappop(q)  
            
            if current_state in visited:
                continue
            
            if current_state == goal:
                return cost, path + (current_state,)  
            
            visited.add(current_state)
            
            for neighbor in n(current_state):
                heappush(q, (cost + h(neighbor) + 1, cost + 1, neighbor, path + (current_state,)))
        
        return -1, () 
    """
    
    def h(state):
        return sum(abs((v-1) % 3 - i % 3) + abs((v-1) // 3 - i // 3) for i, v in enumerate(state) if v != 0)

    def n(state):
        idx = state.index(0)  
        neighbors = []
        for d in (-1, 1, -3, 3):  
            new_idx = idx + d
            if 0 <= new_idx < 9:

                if (d == -1 and idx % 3 == 0) or (d == 1 and idx % 3 == 2):
                    continue

                new_state = list(state)
                new_state[idx], new_state[new_idx] = new_state[new_idx], new_state[idx]
                neighbors.append(tuple(new_state))
        return neighbors


    q, visited = [(h(start), 0, start, ())], set()

    while q:
        _, cost, current_state, path = heappop(q)  
        
        if current_state in visited:
            continue
        
        if current_state == goal:
            return cost, path + (current_state,)  
        
        visited.add(current_state)
        
        for neighbor in n(current_state):
            heappush(q, (cost + h(neighbor) + 1, cost + 1, neighbor, path + (current_state,)))
    
    return -1, () 


